﻿using System;
using System.Diagnostics;
using System.Threading.Tasks;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Mapping;
using com.orbitgis.tool.message;
using orbitgt.mapping3d.sdk.viewer;

namespace plugin_3dm_arcgispro
{
    /// <summary>
    /// Class UpdateFootprint manages the updates to the footprint layer
    /// </summary>
    class UpdateFootprint
    {
        private MapObjectLayer layer;
        private PluginArcGIS viewerHost;
        private long mapCanvasUpdateTime;
        private long lastRequestTime;
        private bool isUpdating;
        private AList<MapObject> mapObjects;

        /// <summary>
        /// Create a new updater
        /// </summary>
        internal UpdateFootprint(MapObjectLayer layer, PluginArcGIS viewerHost)
        {
            this.layer = layer;
            this.viewerHost = viewerHost;
            this.mapCanvasUpdateTime = 0;
            this.lastRequestTime = 0;
            this.isUpdating = false;
            this.mapObjects = null;
        }

        /// <summary>
        /// Called when the mapcanvas has changed and the footprints need to be reloaded
        /// </summary>
        internal void updatedMapCanvas()
        {
            this.mapCanvasUpdateTime = DateTime.Now.Ticks;
        }

        /// <summary>
        /// Check for updates
        /// </summary>
        internal void checkForUpdates()
        {
            // already updating?
            if (this.isUpdating) return;
            // already up-to-date?
            if (this.lastRequestTime > this.mapCanvasUpdateTime) return;
            long currentTime = DateTime.Now.Ticks;
            // throttle the rate of update requests
            if ((this.lastRequestTime > 0) && (currentTime < this.lastRequestTime + 1000 * TimeSpan.TicksPerMillisecond)) return;
            // throttle the rate of mapcanvas changes (fast zoomin operation)
            if ((this.mapCanvasUpdateTime > 0) && (currentTime < this.mapCanvasUpdateTime + 1000 * TimeSpan.TicksPerMillisecond)) return;
            // start the update task (to free the calling timer thread)
            this.isUpdating = true;
            this.lastRequestTime = currentTime;
            Task.Run(() => this.update());
        }

        /// <summary>
        /// Update
        /// </summary>
        private void update()
        {
            try
            {
                Message.WriteLine("Getting map extent");
                MapView mapView = this.viewerHost.getMapView();
                if (mapView == null) return;
                Envelope mapExtent = mapView.Extent;
                if (mapExtent == null) return;
                SpatialReference mapCRS = mapExtent.SpatialReference;
                Message.WriteLine("The map CRS is " + mapCRS.Wkid);
                Message.WriteLine("The map X range is " + mapExtent.XMin + " to " + mapExtent.XMax);
                Message.WriteLine("The map Y range is " + mapExtent.YMin + " to " + mapExtent.YMax);
                if (mapCRS == null) return;
                double mapScale = mapView.Camera.Scale; // denominator
                MapPoint p0 = GeometryEngine.Instance.Project(MapPointBuilder.CreateMapPoint(mapExtent.XMin, mapExtent.YMin, mapCRS), MapObjects.VIEWER_CRS) as MapPoint;
                MapPoint p1 = GeometryEngine.Instance.Project(MapPointBuilder.CreateMapPoint(mapExtent.XMax, mapExtent.YMin, mapCRS), MapObjects.VIEWER_CRS) as MapPoint;
                MapPoint p2 = GeometryEngine.Instance.Project(MapPointBuilder.CreateMapPoint(mapExtent.XMin, mapExtent.YMax, mapCRS), MapObjects.VIEWER_CRS) as MapPoint;
                MapPoint p3 = GeometryEngine.Instance.Project(MapPointBuilder.CreateMapPoint(mapExtent.XMax, mapExtent.YMax, mapCRS), MapObjects.VIEWER_CRS) as MapPoint;
                Bounds worldExtent = new Bounds(""+ MapObjects.VIEWER_CRS_CODE);
                worldExtent.addXY(p0.X, p0.Y);
                worldExtent.addXY(p1.X, p1.Y);
                worldExtent.addXY(p2.X, p2.Y);
                worldExtent.addXY(p3.X, p3.Y);
                Message.WriteLine("Requesting streaming footprints: bbox " + worldExtent + " scale " + mapScale);
                Promise<AList<MapObject>> mapObjectsPromise = this.viewerHost.getViewer().getFootprints2D(worldExtent, mapScale);
                Message.WriteLine("Awaiting footprints with promise id " + mapObjectsPromise.getId());
                this.mapObjects = mapObjectsPromise.waitFor();
                Message.WriteLine("Got " + this.mapObjects.size() + " mapobjects");
                this.isUpdating = false;
                /* Update the layer */
                this.layer.update(this.mapObjects);
                this.viewerHost.RequestMapUpdate("updated footprints");
                return;
            }
            catch (Exception e)
            {
                /* Log the exception */
                Message.WriteLine("Error while updating footprints: " + e);
                this.isUpdating = false;
                return;
            }
        }
    }
}
